#include "stdafx.h"

#include <cstdlib> // For some useful functions such as atexit :)
#include "SDL/SDL-1.2.14 sdk/include/SDL.h" // main SDL header
#include "SDL/SDL-1.2.14 sdk/include/SDL_opengl.h" 
#include "SDL/SDL-1.2.14 sdk/include/SDL_syswm.h"

//#include "SceneCollections/SceneCollections_SparsePtrGrid.h"

#include "Gfx/Gfx_Font.h"
#include "Gfx/Gfx_Funcs.h"
#include "Gfx/Gfx_Viewport.h"
#include "Input/Input_InputMngr.h"
#include "Input/Input_Keys.h"
#include "Input/Input_Mouse.h"
#include "Misc/Debug_GfxStream.h"
#include "Noise/Noise_Random.h"
#include "Plot/Plot_Prims.h"
#include "Profile/Profile_ProfileScope.h"

#include "../Layers/Layers_All.h"
#include "../Session/Session_Settings.h"

#include "App_Main.h"

namespace App
{
	Main::Main (int argc, _TCHAR* argv[])
		: isRunning (true),
		Surf_Display (NULL),
		font (NULL),
		inputMngr (NULL),
		layers (NULL),
		timePrev (0),
		timeStepAccumulator (0),
		settingsFilename (NULL)
	{
		int tmpFlag = _CrtSetDbgFlag (_CRTDBG_REPORT_FLAG);
		tmpFlag |= _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF;
		_CrtSetDbgFlag (tmpFlag);

		//_CrtSetAllocHook (MyAllocFunc);
		//_CrtSetBreakAlloc (5931);

		ParseCommandLineArgs (argc, argv);

		inputMngr = aah_new Input::InputMngr (Int2 (1280, 720));
	}

	Main::~Main ()
	{
		delete [] settingsFilename;

		printf ("debug complete\n");

		//_CrtDumpMemoryLeaks  ();
	}

	int Main::OnExecute ()
	{
		if (OnInit () == false) 
		{
			return -1;
		}
		Profile::ProfileApp::ReportToScreen ("init timings");

		SDL_Event Event;

		while (isRunning)
		{
			while (SDL_PollEvent (&Event)) 
			{
				OnEvent (&Event);
			}

			OnLoop ();
			OnRender ();

			if (layers->isQuitRequested)
			{
				isRunning = false;
			}
		}

		OnCleanup ();

		return 0;
	}

	bool Main::OnInit ()
	{
		Profile::ProfileApp::Init ();
		Profile::ProfileApp::RegisterThread ("client main thread");

		//PROFILE_RECORD_SCOPE ("Main::OnInit");

		if (SDL_Init (SDL_INIT_EVERYTHING) < 0) 
		{
			return false;
		}

		SDL_WM_SetCaption ("Tiny Galaxy. Built " __DATE__ " " __TIME__, NULL);

		SDL_EnableUNICODE (1);
		int result = SDL_EnableKeyRepeat (SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
		assert (result == 0);

		SDL_GL_SetAttribute (SDL_GL_SWAP_CONTROL, 1 );
		SDL_GL_SetAttribute (SDL_GL_DOUBLEBUFFER, 1 ); // *new*

		int ogl = SDL_GL_LoadLibrary (NULL);
		if (ogl != 0)
		{
			return false;
		}

		Surf_Display = SDL_SetVideoMode (inputMngr->GetViewport ()->size.x, inputMngr->GetViewport ()->size.y, 32, SDL_OPENGL | SDL_RESIZABLE);

		if (Surf_Display == NULL) 
		{
			return false;
		}

		SDL_EventState (SDL_MOUSEMOTION, SDL_IGNORE);
		SDL_EventState (SDL_MOUSEMOTION, SDL_DISABLE);

		Gfx::LoadOpenGlFunctions ();

		Debug_Init (0, aah_new DebugStream_Console ());
		//Debug_WriteLine (0, "Ludum Dare 23 (" __DATE__ ", " __TIME__ ")");
		//Debug_WriteLine (0, "www.aahgames.com");
		//Debug_WriteLine (0, "");

		font = aah_new Gfx::Font ("fonts/nokia_medium.glf");
		Debug_Init (1, aah_new Debug::DebugStream_Gfx (font, 20, 20, 20));

		Session::Settings::Load (settingsFilename);

		Plot::Prims_Init (inputMngr);

		layers = aah_new Layers::All (inputMngr);
		
		return true;
	}

	void Main::OnEvent (SDL_Event * Event) 
	{
		//PROFILE_RECORD_SCOPE ("Main::OnEvent");

		SdlEvent::OnEvent (Event);
	}

	void Main::OnLoop ()
	{
		int updatesPerRender = 0;
		const RakNet::Time unitsPerSecond = RakNet::GetTimeTicksPerSecond ();
		const RakNet::Time unitsPerTick = unitsPerSecond / 60LL;

		RakNet::Time timeNow = RakNet::GetTime ();

		if (timePrev != 0LL)
		{
			int updatesMax = 4;
			int updatesCount = 0;

			timeStepAccumulator += timeNow - timePrev;
			while (timeStepAccumulator >= unitsPerTick)
			{
				if (updatesCount < updatesMax)
				{
					{
						//PROFILE_RECORD_SCOPE ("Main::OnLoop single update");

						const float timeStepInSeconds = (float) unitsPerTick / (float) unitsPerSecond;

						inputMngr->Update ();
						layers->Update (timeNow, timeStepInSeconds);
						//timeNow += timePerTickMs;
					}

					Profile::ProfileApp::Update (); // do this outside the loop
				}
				++updatesCount;
				timeStepAccumulator -= unitsPerTick;
				++updatesPerRender;
			}
		}
		timePrev = timeNow;

		const float weight = (float) timeStepAccumulator / (float) unitsPerTick;
		layers->UpdateFractionBlend (timeNow, weight);
	}

	void Main::OnRender () 
	{
		{
			//PROFILE_RECORD_SCOPE ("OnRender", 128, 128, 0);

			GLenum error = glGetError (); assert (error == GL_NO_ERROR);

			glViewport (0, 0, (GLsizei) inputMngr->GetViewport ()->size.x, (GLsizei) inputMngr->GetViewport ()->size.y);

			error = glGetError (); assert (error == GL_NO_ERROR);

			layers->Render ();
			
			glColor3f (1.0f, 1.0f, 1.0f);

			Plot::Prims_BatchBegin (Plot::PT_XYRGB); // dodgy method of resetting renderstates and viewport
			Plot::Prims_BatchEnd ();

			font->Begin ();
			Debug_Flush (1);
			font->End ();
		}

		Plot::Plot_Profiler (inputMngr->GetViewport ()->size);

		SDL_GL_SwapBuffers();

		Debug_Flush (0);
	}

	void Main::OnCleanup ()
	{
		delete layers;

		delete font;

		delete inputMngr;

		Session::Settings::Unload ();

		SDL_FreeSurface (Surf_Display);
		SDL_Quit ();

		Profile::ProfileApp::Exit ();

		Debug_Exit ();
	}

	void Main::OnExit()
	{
		//comms->PauseUntilAsyncsKilled ();
		//server->PauseUntilAsyncsKilled ();
		//client->PauseUntilAsyncsKilled ();

		isRunning = false;
	}

	void Main::OnKeyDown (SDLKey sym, SDLMod mod, Uint16 unicode)
	{
		inputMngr->GetKeys ()->RecordKeyPressed (sym, mod, unicode);
	}

	void Main::OnKeyUp (SDLKey sym, SDLMod mod, Uint16 unicode)
	{
		inputMngr->GetKeys ()->RecordKeyReleased (sym, mod, unicode);
	}

	void Main::OnMouseMove (int x, int y, int dx, int dy, bool l, bool r, bool m)
	{
		// unused as we poll windows directly for its position - due to annoying focus bug in SDL
		//inputMngr->OnMouseMove (x, y, dx, dy, l, r, m);
	}

	void Main::OnLButtonDown (int mX, int mY)
	{
		inputMngr->GetMouse ()->ReportMouseButton (0, true);
	}

	void Main::OnLButtonUp (int mX, int mY)
	{
		inputMngr->GetMouse ()->ReportMouseButton (0, false);
	}

	void Main::OnRButtonDown (int mX, int mY)
	{
		inputMngr->GetMouse ()->ReportMouseButton (1, true);
	}

	void Main::OnRButtonUp (int mX, int mY)
	{
		inputMngr->GetMouse ()->ReportMouseButton (1, false);
	}

	void Main::OnMButtonDown (int mX, int mY)
	{
		inputMngr->GetMouse ()->ReportMouseButton (2, true);
	}

	void Main::OnMButtonUp (int mX, int mY)
	{
		inputMngr->GetMouse ()->ReportMouseButton (2, false);
	}

	void Main::OnMWheelDown (bool isUpClick)
	{
		inputMngr->GetMouse ()->ReportWheel (isUpClick);
	}

	void Main::OnMouseEnterWindow ()
	{
	}

	void Main::OnMouseLeaveWindow ()
	{
		//if (inputMngr->GetMouse ()->)
		//{
		//	inputMngr->OnRequestExclusive (false); 
		//}
	}

	void Main::OnInputFocusGained ()
	{
		inputMngr->GetMouse ()->SetInputFocus (true);
	}

	void Main::OnInputFocusLost ()
	{
		inputMngr->GetMouse ()->SetInputFocus (false);
	}

	void Main::OnVisible ()
	{
	}

	void Main::OnInvisible ()
	{
	}

	void Main::OnResize (int w,int h)
	{
		inputMngr->GetViewport ()->size = Int2 (w, h);
	}

	void Main::ParseCommandLineArgs (int argc, _TCHAR* argv[])
	{
		for (int current = 1; current < argc; ++current)
		{
			const char * arg = argv [current];
			if (strcmp (arg, "-s") == 0)
			{
				settingsFilename = strdup (argv [++current]);
			}
		}

		if (settingsFilename == NULL)
		{
			settingsFilename = strdup ("scripts/settings_client.lua");
		}
	}
}
