#include "stdafx.h"

#include "Gfx/Gfx_Camera.h"
#include "Gfx/Gfx_Font.h"
#include "Gfx/Gfx_Viewport.h"
#include "Input/Input_InputMngr.h"
#include "Input/Input_Keys.h"
#include "Input/Input_Mouse.h"
#include "Particles/Particles_BillboardMngr.h"
#include "Profile/Profile_ProfileScope.h"
#include "SceneGfx/SceneGfx_Column.h"
#include "SceneGfx/SceneGfx_Room.h"
#include "SceneGfx/SceneGfx_World.h"
#include "SceneHelper/SceneHelper_SimMngr.h"
#include "SceneSim/SceneSim_Defs.h"
#include "SceneSim/SceneSim_Column.h"
#include "SceneSim/SceneSim_Room.h"
#include "SceneSim/SceneSim_World.h"

#include "../Ents/Ents_Player.h"
#include "../Layers/Layers_All.h"
#include "../Layers/Layers_RenderStates.h"
#include "../Tiny/Tiny_TinyWorld.h"
#include "../Tiny/Tiny_TinyWorlds.h"

#include "LayerPlaying_Playing.h"
#include "LayerPlaying_PlayingDialog.h"
#include "LayerPlaying_PlayingMenus.h"
#include "LayerPlaying_worldSyncMngr.h"

namespace LayerPlaying
{
	Playing::Playing (Layers::All * all)
		: isRunningRequested (false),
		state (S_NONE),
		isRunning (false),
		all (all),
		tinyWorlds (NULL)
	{
		font = aah_new Gfx::Font ("fonts/nokia_medium.glf");

		tinyWorlds = NULL;
		update = NULL;
		worldSyncMngr = NULL;
		gfxWorld = NULL;
		simWorld = NULL;

		columnMask = NULL;
		homeOnlyMask = NULL;
	}

	Playing::~Playing ()
	{
		DeleteAll ();
	}

	void Playing::DeleteAll ()
	{
		delete columnMask;
		columnMask = NULL;

		delete homeOnlyMask;
		homeOnlyMask = NULL;

		delete tinyWorlds;
		tinyWorlds = NULL;

		if (update != NULL)
		{
			update->tinyWorlds = NULL;

			delete update->billboards;
			update->billboards = NULL;
		}

		delete worldSyncMngr; 
		worldSyncMngr = NULL;

		delete gfxWorld; 
		gfxWorld = NULL;

		delete simWorld; 
		simWorld = NULL;

		delete font;
		font = NULL;
	}

	void Playing::Preupdate (ClientEnts::UpdateData * update)
	{
		if (isRunningRequested && !isRunning)
		{
			assert (!isRunning);
			
			state = S_LOADINGSIM;
			isGraphicsComplete = false;

			update->inputMngr->GetMouse ()->SetIsFpsMouse (true);

			update->sim = simWorld = aah_new SceneSim::World ();
			update->renderer = gfxWorld = aah_new SceneGfx::World ();
			update->worldSyncMngr = worldSyncMngr = aah_new WorldSyncMngr (simWorld, gfxWorld, false);
			update->billboards = aah_new Particles::BillboardMngr (300);

			Debug_WriteLine (1, "Loading Tiny Galaxy...");

			columnMask = aah_new Ld::ColumnMask ();
			homeOnlyMask = aah_new Ld::ColumnMask ();
			homeOnlyMask->Set (false);
			homeOnlyMask->use [0][0] = true;

			tinyWorlds = aah_new Tiny::TinyWorlds (gfxWorld);
			tinyWorlds->SetMask (columnMask);

			update->tinyWorlds = tinyWorlds;
		}
		else if (!isRunningRequested && isRunning)
		{
			assert (isRunning);

			DeleteAll ();
		}
		isRunning = isRunningRequested;
	}

	void Playing::Update (ClientEnts::UpdateData * update)
	{
		if (isRunning)
		{
			switch (state)
			{
			case S_NONE:
				assert (false);
				break;

			case S_LOADINGSIM:
				worldSyncMngr->Update ();
				if (worldSyncMngr->IsSimReady)
				{
					all->playingDialog->Reset ();
					InitEnts (update);
					UpdateWorld (update, false);
					Debug_WriteLine (1, "Generating Tiny Galaxy...");
					state = S_GENERATINGGFX;
				}
				break;

			case S_GENERATINGGFX:
				UpdateWorld (update, false);
				worldSyncMngr->Update ();
				if (worldSyncMngr->IsGfxReady)
				{
					//Debug_WriteLine (1, "GRAPHICS READY");
					state = S_PLAYING;
				}
				break;

			case S_PLAYING:
				worldSyncMngr->Update ();
				UpdateWorld (update, true);
				break;
			}

			const bool allowUserInput = !all->playingMenus->GetIsRunning ();

			if (allowUserInput)
			{
				if (update->inputMngr->GetMouse ()->GetIsFpsMouse ())
				{
					if (update->inputMngr->GetKeys ()->IsPressed (SDLK_ESCAPE))
					{
						all->playingMenus->isRunningRequested = true;
					}
				}
			}
		}
	}

	void Playing::InitEnts (ClientEnts::UpdateData * update)
	{
		update->player = aah_new Ents::Player (false, update->renderer->GetPrimaryRoom ()->GetCellMesh (), all);
		update->player->Init (Int3 (1, 0, 1), Vec3 (1.0f, 32.0f, 1.0f));
	}

	void Playing::UpdateWorld (ClientEnts::UpdateData * update, bool isLoadingFinished)
	{
		this->update = update;

		update->columnMask = columnMask;

		update->billboards->Update (update->timeStep);

		Ents::Player::STATES oldState = update->player->state;
		update->player->Update (update, !(all->playingMenus->GetIsRunning () || all->playingDialog->GetIsRunning ()));
		Ents::Player::STATES newState = update->player->state;

		Int3 home = update->player->homeOnlyColumnId;

		if (oldState == Ents::Player::S_PLAYING && newState == Ents::Player::S_FLYING)
		{
			worldSyncMngr->PauseUntilSynced ();
			SwapSimAndGfxColumns (home, Int3 (0, 0, 0));
			columnMask->use [home.x][home.z] = false;
		}
		else if (oldState == Ents::Player::S_FLYING && newState == Ents::Player::S_PLAYING)
		{
			worldSyncMngr->PauseUntilSynced ();
			SwapSimAndGfxColumns (home, Int3 (0, 0, 0));
			tinyWorlds->worlds [0]->source [0] = home;
			tinyWorlds->SetMask (columnMask);
		}

		tinyWorlds->Update ();

		if (isLoadingFinished)
		{
			Debug_Clear (1);

			//Debug_WriteLine (1, "GRAPHICS READY");

			Ents::Comps::Coord c;
			update->player->GetEyeCoord (&c);

			const Tiny::TinyWorld * tworld = tinyWorlds->FindWorld (c.GetColumnId ());
			if (tworld != NULL)
			{
				Debug_WriteLine (1, tworld->title);
				char buffer [256];
				if (update->player->artifactCount == 1)
				{
					sprintf (buffer, "%d artifact of 6 found", update->player->artifactCount);
				}
				else
				{
					sprintf (buffer, "%d artifacts of 6 found", update->player->artifactCount);
				}
				Debug_WriteLine (1, buffer);
			}


		}


		update->columnMask = NULL;
	}

	void Playing::UpdateFractionBlend (ClientEnts::UpdateData * update, float weight)
	{
		if (isRunning)
		{
			//update->you->UpdateFractionBlend (update, weight);
		}
	}

	void Playing::Render ()
	{
		if (isRunning)
		{
			glClearColor (0.0f, 0.0f, 0.0f, 1.0f);
			glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

			Layers::SetRenderStates_3d ();

			switch (state)
			{
			case S_NONE:
			case S_LOADINGSIM:
				break;

			case S_GENERATINGGFX:
			case S_PLAYING:
				{
					Int3 cameraColumnId (0, 0, 0);
					Vec3 cameraColResPos (0.0f, 0.0f, 0.0f);

					Gfx::Camera * camera = update->inputMngr->GetCamera ();
					camera->viewportSize = update->inputMngr->GetViewport ()->size;
					update->player->CalcCamera (camera, &cameraColumnId, &cameraColResPos);

					glMatrixMode (GL_PROJECTION);
					glLoadMatrixf ((GLfloat *) &camera->matProjection);
					glMatrixMode (GL_MODELVIEW);
					glLoadMatrixf ((GLfloat *) &camera->matView);

					//skydome->Render (cameraColResPos);
					//glCullFace (GL_BACK);

					//gfxWorld->GetPrimaryRoom ()->Render_Begin (0);
					//tinyWorlds->Render (cameraColumnId, cameraColResPos);
					//gfxWorld->GetPrimaryRoom ()->Render_End (0);

					gfxWorld->Render (0, columnMask, cameraColumnId, cameraColResPos, 1000.0f, Int3 (0, 0, 0));

					switch (update->player->state)
					{
					case Ents::Player::S_PLAYING: break;

					case Ents::Player::S_FLYING:
						{
							Ents::Player * p = update->player;
							gfxWorld->Render (0, homeOnlyMask, cameraColumnId, cameraColResPos, 1000.0f, -p->homeOnlyColumnId);
						}
						break;
					}

					glDisable (GL_CULL_FACE); 

					if (update->player != NULL)
					{
						Gfx::Camera * camera = update->inputMngr->GetCamera ();

						const Mat44 & m = camera->matView;
						Vec3 right (m [0].x, m [1].x, m [2].x);
						Vec3 up (m [0].y, m [1].y, m [2].y);
						Vec3 in (m [0].z, m [1].z, m [2].z);

						update->billboards->Render (cameraColumnId, right, up, in);
					}

					update->player->Render (camera);
				}
				break;
			}
		}
	}

	bool Playing::GetIsRunning () const
	{
		return isRunning;
	}

	void Playing::SwapSimAndGfxColumns (Int3 a, Int3 b)
	{
		if (a == b)
		{
			return;
		}

		SceneGfx::Room * gfxRoom = gfxWorld->GetPrimaryRoom ();
		gfxRoom->GetColumn (a)->SwapWith (gfxRoom->GetColumn (b));
	}
}
