#include "stdafx.h"

#include "GfxFactory/GfxFactory_GfxFactory.h"
#include "GfxFactory/GfxFactory_JobApplyCsgSim.h"
#include "SceneChange/Cmds_ApplyCsg.h"
#include "SceneFile/SceneFile_SimFactory.h"
#include "SceneGfx/SceneGfx_Column.h"
#include "SceneGfx/SceneGfx_Room.h"
#include "SceneGfx/SceneGfx_World.h"
#include "SceneSim/SceneSim_Room.h"
#include "SceneSim/SceneSim_World.h"

#include "../Session/Session_Settings.h"

#include "LayerPlaying_WorldSyncMngr.h"

namespace LayerPlaying
{
	WorldSyncMngr::WorldSyncMngr (SceneSim::World * simWorld, SceneGfx::World * gfxWorld, bool isEditMode)
		: isEditMode (isEditMode),
		simWorld (simWorld),
		gfxWorld (gfxWorld)
	{
		simFactory = aah_new SceneFile::SimFactory (&Session::Settings::Get ().worldInitParams, simWorld, this, isEditMode);
		gfxFactory = aah_new GfxFactory::GfxFactory (simWorld, gfxWorld);

		IsSimReady = false;
		IsGfxReady = false;
		gfxColumnInitCount = 0;
	}

	WorldSyncMngr::~WorldSyncMngr ()
	{
		if (isEditMode)
		{
			SaveAndUnload ();
		}
	}

	void WorldSyncMngr::Update ()
	{
		if (!IsSimReady || !IsGfxReady)
		{
			InitialiseWorld ();
		}

		simFactory->Update ();
		gfxFactory->Update ();
	}

	void WorldSyncMngr::ApplyCsg (RakNet::Time * time, TextRef roomId, Int3 columnId, Int3 cellId, float radius, int cellType, int cellData)
	{
		GfxFactory::JobApplyCsgSim * job = aah_new GfxFactory::JobApplyCsgSim (gfxFactory);
		job->applied = *time;
		job->roomId = roomId;
		job->columnId = columnId;
		job->cellId = cellId;
		job->radius = radius;
		job->cellType = cellType;
		job->cellData = cellData;

		job->AddToQueue (Thread::JB_DEFAULT);
	}

	void WorldSyncMngr::OnWorldAdded (SceneSim::World * world)
	{
		gfxFactory->OnSimWorldAdded (world);
	}

	void WorldSyncMngr::OnWorldRemoved (SceneSim::World * world) 
	{
		assert (false);
	}

	void WorldSyncMngr::OnRoomAdded (SceneSim::Room * room) 
	{
		gfxFactory->OnSimRoomAdded (room);
	}

	void WorldSyncMngr::OnRoomRemoved (SceneSim::Room * room) 
	{
		assert (false);
	}

	void WorldSyncMngr::OnColumnAdded (SceneSim::Column * column) 
	{
		gfxFactory->OnSimColumnAdded (column);
	}

	void WorldSyncMngr::OnColumnRemoved (SceneSim::Column * column) 
	{
		// ignore this for now - only removing now on level save when quitting
	}

	void WorldSyncMngr::OnTerrainCsgApplied (SceneSim::Column * column, Int3 cellId, float radius, int cellType) 
	{
	}

	void WorldSyncMngr::InitialiseWorld ()
	{
		if (!IsSimReady)
		{
			IsSimReady = GenerateInitialSim ();
		}

		if (IsSimReady && !IsGfxReady)
		{
			IsGfxReady = GenerateInitialGfx ();
		}
	}

	bool WorldSyncMngr::GenerateInitialSim ()
	{
		bool anyLoadRequested = true;

		if (simWorld->loadState == SceneSim::LS_NONE)
		{
			simFactory->RequestWorldLoad ();
		}
		
		if (simWorld->loadState == SceneSim::LS_LOADCOMPLETE)
		{
			if (!simWorld->IsRoomLoaded ("hub"))
			{
				simFactory->RequestRoomLoad ("hub");
			}
			else
			{
				anyLoadRequested = false;

				SceneSim::Room * room = simWorld->GetRoom ("hub");
				for (int idx = 0; idx < SceneSim::FixedSize * SceneSim::FixedSize; ++idx)
				{
					Int3 columnId (idx % SceneSim::FixedSize, 0, idx / SceneSim::FixedSize);
					if (room->GetColumn (columnId) == NULL)
					{
						if (!simFactory->IsColumnRequested (room, columnId))
						{
							simFactory->RequestColumnLoad (room, columnId);
						}
						anyLoadRequested = true;
					}
				}
			}
		}

		return (!anyLoadRequested);
	}

	bool WorldSyncMngr::GenerateInitialGfx ()
	{
		bool anyLoadRequested = (gfxFactory->GetJobCount () > 0);

		if (!anyLoadRequested)
		{
			if (gfxColumnInitCount < SceneSim::FixedSize * SceneSim::FixedSize)
			{
				Int3 columnId (gfxColumnInitCount % SceneSim::FixedSize, 0, gfxColumnInitCount / SceneSim::FixedSize);
				SceneGfx::Room * gfxRoom = gfxWorld->GetPrimaryRoom ();
				gfxFactory->RequestGfxColumnMeshRefresh (columnId);

				++gfxColumnInitCount;
				anyLoadRequested = true;
			}
		}

		return (!anyLoadRequested);
	}

	void WorldSyncMngr::PauseUntilSynced ()
	{
		while (gfxFactory->GetJobCount () > 0 && simFactory->GetJobCount () > 0)
		{
			simFactory->Update ();
			gfxFactory->Update ();
		}
	}

	void WorldSyncMngr::SaveAndUnload ()
	{
		if (gfxFactory != NULL)
		{
			delete gfxFactory;
			gfxFactory = NULL;
		}

		if (simFactory != NULL)
		{
			simFactory->AbortAsyncAndSave ();
			delete simFactory; 
			simFactory = NULL;
		}
	}
}

