#include "stdafx.h"

#include "Gfx/Gfx_Font.h"
#include "Gfx/Gfx_Viewport.h"
#include "Input/Input_InputMngr.h"
#include "Input/Input_Keys.h"
#include "Input/Input_Mouse.h"
#include "Plot/Plot_Prims.h"

#include "Controls_Base.h"
#include "Controls_Button.h"
#include "Menus_Draw.h"
#include "Menus_Page.h"
#include "Menus_Palette.h"

namespace Menus
{
	Page::Page (Input::InputMngr * inputMngr, Gfx::Font * font, TextRef id)
		: isActive (false),
		inputMngr (inputMngr),
		font (font),
		Id (id),
		pageSize (0, 0),
		toCentre (0, 0),
		focusedControl (-1),
		highlightedControl (-1),
		isMouseHighlighting (true),
		buttonPressed (NULL),
		buttonPressedTimer (0),
		onEntryCallback (NULL),
		onEntryData (NULL),
		onExitCallback (NULL),
		onExitData (NULL),
		note (NULL)
	{
		assert (inputMngr != NULL);
		assert (font != NULL);
	}

	Page::~Page ()
	{
		RemoveControls ();

		font = NULL;
		inputMngr = NULL;
	}

	Menus::Controls::Button * Page::Add (Menus::Controls::Button * control)
	{
		buttons.push_back (control);
		controls.push_back (control);
		control->page = this;
		return control;
	}

	Menus::Controls::Base * Page::Add (Menus::Controls::Base * control)
	{
		controls.push_back (control);
		control->page = this;
		return control;
	}

	void Page::MarkBottomRight (Int2 pageSize)
	{
		assert (pageSize.x > 0 && pageSize.y > 0);
		this->pageSize = pageSize;
	}

	void Page::SetOnEntry (PageCallback callback, void * data)
	{
		assert (callback != NULL);
		
		onEntryCallback = callback;
		onEntryData = data;
	}

	void Page::SetOnExit (PageCallback callback, void * data)
	{
		assert (callback != NULL);
		
		onExitCallback = callback;
		onExitData = data;
	}

	void Page::RemoveControls ()
	{
		for (int idx = 0; idx < controls.size (); ++idx)
		{
			delete controls [idx];
		}
		controls.clear ();
		buttons.clear ();

		buttonPressed = NULL;
	}

	void Page::Init (bool resetPage)
	{
		isMouseHighlighting = true;

		if (resetPage)
		{
			if (focusedControl != -1)
			{
				controls [focusedControl]->SetIsFocused (false);
			}
			focusedControl = FindNextHighlightableControl (0, 1);
			if (focusedControl != -1)
			{
				controls [focusedControl]->SetIsFocused (true);
			}
		}

		if (highlightedControl != -1)
		{
			controls [highlightedControl]->SetIsHighlighted (false);
		}
		highlightedControl = -1;

		buttonPressed = NULL;
		buttonPressedTimer = 0;

		for (int idx = 0; idx < (signed) buttons.size (); ++idx)
		{
			buttons [idx]->wasPressed = false;
		}

		Int2 viewport (inputMngr->GetViewport ()->size);
		toCentre = Int2 ((viewport.x - pageSize.x) / 2, (viewport.y - pageSize.y) / 2);
	}

	void Page::SetActive (bool isActive)
	{
		this->isActive = isActive;
		
		if (isActive)
		{
			if (onEntryCallback != NULL)
			{
				onEntryCallback (this, onEntryData);
			}
		}
		else
		{
			if (onExitCallback != NULL)
			{
				onExitCallback (this, onExitData);
			}
		}
	}

	void Page::Update ()
	{
		if (!isActive)
		{
			return;
		}

		if (buttonPressed == NULL)
		{
			Int2 viewport (inputMngr->GetViewport ()->size);
			toCentre = Int2 ((viewport.x - pageSize.x) / 2, (viewport.y - pageSize.y) / 2);

			const Input::Mouse * mouse = inputMngr->GetMouse ();
			assert (!mouse->GetIsFpsMouse ());

			Int2 cursor (mouse->GetAbsolute ());
			cursor -= toCentre;

			const bool hasMouseMoved = (oldMouse != cursor);
			if (hasMouseMoved)
			{
				isMouseHighlighting = true;
			}

			if (isMouseHighlighting)
			{
				int nextHighlightedControl = -1;
				if (mouse->GetIsAbsoluteValid ())
				{
					for (int idx = 0; idx < (signed) controls.size (); ++idx)
					{			
						if (controls [idx]->IsHighlighted (cursor))
						{
							nextHighlightedControl = idx;
							break;
						}
					}
				}

				ChangeHighlighted (nextHighlightedControl);

				if (mouse->GetIsAbsoluteValid () && mouse->Left.IsJustOn ())
				{
					ChangeFocused (highlightedControl);
					isMouseHighlighting = false;
				}
			}

			if (focusedControl != -1)
			{
				if (focusedControl == NULL || !controls [focusedControl]->IsUsingAllKeyInput ())
				{
					const Input::Keys * keys = inputMngr->GetKeys ();
					if (keys->IsPressed (SDLK_DOWN))
					{
						int nextIdx = FindNextHighlightableControl (focusedControl, +1);
						if (nextIdx != -1)
						{
							ChangeHighlighted (nextIdx);
							ChangeFocused (nextIdx);
							isMouseHighlighting = false;
						}
					}
					else if (keys->IsPressed (SDLK_UP))
					{
						int nextIdx = FindNextHighlightableControl (focusedControl, -1);
						if (nextIdx != -1)
						{
							ChangeHighlighted (nextIdx);
							ChangeFocused (nextIdx);
							isMouseHighlighting = false;
						}
					}
				}
			}

			for (int idx = 0; idx < (signed) controls.size (); ++idx)
			{
				controls [idx]->Update (inputMngr);
			}

			for (int idx = 0; idx < (signed) buttons.size (); ++idx)
			{
				if (buttons [idx]->wasPressed)
				{
					buttonPressed = buttons [idx];
					break;
				}
			}

			oldMouse = cursor;
		}
		else
		{
			++buttonPressedTimer;
			if (buttonPressedTimer == 10)
			{
				if (!buttonPressed->Executed ())
				{
					buttonPressedTimer = 0;
					buttonPressed = NULL;
				}
			}
		}
	}

	void Page::Render ()
	{
		if (!isActive)
		{
			return;
		}

		for (int passIdx = 0; passIdx < 2; ++passIdx)
		{
			if (passIdx == 0)
			{
				Plot::Prims_BatchBegin (Plot::PT_XYRGB);

				//error = glGetError (); assert (error == GL_NO_ERROR);

				const Palette * p = Palette_Get (false);

				int thickness = 4;
				Int2 pos (0, toCentre.y - thickness);
				Int2 size (inputMngr->GetViewport ()->size.x, pageSize.y + thickness);
				Draw_Box (pos, size, p->pageBg);

				//error = glGetError (); assert (error == GL_NO_ERROR);
			}
			else
			{
				glDisable (GL_DEPTH_TEST);
				glDisable (GL_ALPHA_TEST);
				glDisable (GL_CULL_FACE); 

				glColor3f (1.0f, 1.0f, 1.0f);

				font->Begin ();
			}

			for (int idx = 0; idx < (signed) controls.size (); ++idx)
			{
				//error = glGetError (); assert (error == GL_NO_ERROR);
				controls [idx]->Render (toCentre, passIdx, font);
				//error = glGetError (); assert (error == GL_NO_ERROR);
			}

			if (passIdx == 0)
			{
				//error = glGetError (); assert (error == GL_NO_ERROR);
				Plot::Prims_BatchEnd ();
			}
			else
			{
				font->End ();
			}
		}
	}

	int Page::GetControlCount () const
	{
		return controls.size ();
	}

	Menus::Controls::Button * Page::GetButtonPressed () const
	{
		if (buttonPressedTimer >= 10 && buttonPressed != NULL)
		{
			return buttonPressed;
		}
		return NULL;
	}

	void Page::ChangeHighlighted (int newIdx)
	{
		if (newIdx != highlightedControl)
		{
			if (highlightedControl != -1)
			{
				controls [highlightedControl]->SetIsHighlighted (false);
			}
			if (newIdx != -1)
			{
				controls [newIdx]->SetIsHighlighted (true);
			}
			highlightedControl = newIdx;
		}
	}

	void Page::ChangeFocused (int newIdx)
	{
		if (newIdx != focusedControl)
		{
			if (focusedControl != -1)
			{
				controls [focusedControl]->SetIsFocused (false);
			}
			if (newIdx != -1)
			{
				controls [newIdx]->SetIsFocused (true);
			}
			focusedControl = newIdx;	
		}
	}

	int Page::FindNextHighlightableControl (int startIdx, int step) const
	{
		startIdx += step;
		while (startIdx >= 0 && startIdx < (signed) controls.size ())
		{
			if (controls [startIdx]->IsHighlightable ())
			{
				return startIdx;
			}
			startIdx += step;
		}
		return -1;
	}
}

