#include "stdafx.h"

#include <vector>

#include "CsgEffect/CsgEffect_ColumnChanges.h"
#include "CsgEffect/CsgEffect_Sphere.h"
#include "SceneCollections/SceneCollections_LocalColumns.h"
#include "SceneSim/SceneSim_Column.h"
#include "SceneSim/SceneSim_ColumnCells.h"
#include "SceneSim/SceneSim_Room.h"
#include "SceneSim/SceneSim_World.h"

#include "../SceneChange/SceneChange_ColumnData.h"

#include "GfxFactory_GfxFactory.h"
#include "GfxFactory_JobApplyCsgGfx.h"
#include "GfxFactory_JobApplyCsgSim.h"

namespace GfxFactory
{
	JobApplyCsgSim::JobApplyCsgSim (GfxFactory * gfxFactory)
		: Job (gfxFactory->jobMngr, true),
		gfxFactory (gfxFactory),
		columnsDirty (NULL),
		columnsEdge (NULL)
	{
	}

	JobApplyCsgSim::~JobApplyCsgSim ()
	{
		delete columnsDirty;

		if (columnsEdge != NULL)
		{
			columnsEdge->DeleteAll ();
			delete columnsEdge;
		}
	}

	Thread::JOBSTATES JobApplyCsgSim::JobDoWork ()
	{
		columnsDirty = aah_new SceneCollections::LocalColumns (columnId, 3);
		std::vector <Int3> dirtyEdges;

		CsgEffect::ColumnChanges changes (cellType, cellData);
		CsgEffect::Sphere * csgEffect = aah_new CsgEffect::Sphere (columnId, cellId, radius);
		const int columnsCount = csgEffect->GetColumnAffectedCount ();

		SceneSim::Room * simRoom = gfxFactory->simWorld->GetRoom (roomId);
		assert (simRoom != NULL);

		const CellBuilder::Mngr * cellBuilder = simRoom->GetCellBuilder ();

		for (int columnIdx = 0; columnIdx < columnsCount; ++columnIdx)
		{
			const Int3 affectedColumnId = csgEffect->GetColumnIdAffected (columnIdx);

			SceneSim::Column * simColumn = simRoom->GetColumn (affectedColumnId);

			if (simColumn != NULL)
			{
				changes.Clear ();
				csgEffect->ApplyChanges (columnIdx, simColumn, cellBuilder, &changes);
				simColumn->SetLastModifiedTime (applied);

				SceneSim::Column * copy = aah_new SceneSim::Column (*simColumn);
				assert (copy->GetRoom () == NULL);
				columnsDirty->AddColumn (copy);

				changes.CalcEdgesTouched ();
				if (changes.edgesTouched [SceneSim::F_N]) {dirtyEdges.push_back (affectedColumnId + Int3 (0, 0, 1));}
				if (changes.edgesTouched [SceneSim::F_E]) {dirtyEdges.push_back (affectedColumnId + Int3 (1, 0, 0));}
				if (changes.edgesTouched [SceneSim::F_S]) {dirtyEdges.push_back (affectedColumnId + Int3 (0, 0, -1));}
				if (changes.edgesTouched [SceneSim::F_W]) {dirtyEdges.push_back (affectedColumnId + Int3 (-1, 0, 0));}
			}
		}

		delete csgEffect;

		// check if any dirtyEdges need adding if they arent already in the columnsDirty collection
		for (int checkIdx = 0; checkIdx < dirtyEdges.size (); ++checkIdx)
		{
			if (columnsDirty->GetColumn (dirtyEdges [checkIdx]) == NULL)
			{
				SceneSim::Column * simColumn = simRoom->GetColumn (dirtyEdges [checkIdx]);
				if (simColumn != NULL)
				{
					SceneSim::Column * copy = aah_new SceneSim::Column (*simColumn);
					columnsDirty->AddColumn (copy);
				}
			}
		}

		// get all the edge columns
		columnsEdge = aah_new SceneCollections::LocalColumns (columnId, 4);
		for (int idx = 0; idx < columnsDirty->GetMaxColumnsCount (); ++idx)
		{
			SceneSim::Column * column = columnsDirty->GetNthColumn (idx);
			if (column != NULL)
			{
				AddEdgeIfEmpty (column->columnId + Int3 (0, 0, 1));
				AddEdgeIfEmpty (column->columnId + Int3 (1, 0, 0));
				AddEdgeIfEmpty (column->columnId + Int3 (0, 0, -1));
				AddEdgeIfEmpty (column->columnId + Int3 (-1, 0, 0));
			}
		}

		// sim updated - now recreate the graphics for it


		return Thread::JS_WORKDONE;
	}

	Thread::JOBSTATES JobApplyCsgSim::JobOnCompleted ()
	{
		SceneSim::Room * simRoom = gfxFactory->simWorld->GetRoom (roomId);
		assert (simRoom != NULL);

		Thread::Job * job = aah_new JobApplyCsgGfx (gfxFactory, &applied, roomId, columnsDirty, columnsEdge);
		job->AddToQueue (Thread::JB_DEFAULT);
		
		delete columnsDirty;
		columnsDirty = NULL;

		delete columnsEdge;
		columnsEdge = NULL;

		return Thread::JS_COMPLETED;
	}

	void JobApplyCsgSim::AddEdgeIfEmpty (Int3 columnId)
	{
		if (columnsDirty->GetColumn (columnId) == NULL && columnsEdge->GetColumn (columnId) == NULL)
		{
			SceneSim::Column * simColumn = gfxFactory->simWorld->GetRoom (roomId)->GetColumn (columnId);
			if (simColumn != NULL)
			{
				columnsEdge->AddColumn (aah_new SceneSim::Column (*simColumn));
			}
		}
	}
}


