#include "stdafx.h"

#include "Lua/Lua_LuaState.h"
#include "Thread/Thread_JobMngr.h"

#include "../Noise/Noise_Perlin.h"
#include "../Noise/Noise_Random.h"
#include "../SceneCollections/SceneCollections_Utils.h"
#include "../SceneFile/Area_AreaMngr.h"
#include "../SceneFile/Factory_ColumnLoadJob.h"
#include "../SceneFile/Factory_ColumnUnloadJob.h"
#include "../SceneFile/SceneFile_Lighting.h"
#include "../SceneSim/SceneSim_Cell.h"
#include "../SceneSim/SceneSim_Column.h"
#include "../SceneSim/SceneSim_ColumnCells.h"
#include "../SceneSim/SceneSim_Room.h"
#include "../SceneSim/SceneSim_World.h"
#include "../SceneSim/SceneSim_WorldInit.h"

#include "Gen00_Room.h"

//#define DEBUG_SIMPLE_COLUMN

namespace SceneGen
{
	namespace Gen00
	{
		const int rockBlocks [][3] =
		{
			{128, 3, 57/*79*/}, // snow
			{2, 3, 1}, // grass
			{12, 24, 4}, // sand
		};

		Gen00_Room::Gen00_Room (SceneSim::Room * room, Thread::JobMngr * const jobMngr, SceneSim::World * world, SceneHelper::SimEvents * events)
			: OperationsRoom (room, jobMngr, world, events),
			randomSeed (0)
		{
		}

		Gen00_Room::~Gen00_Room ()
		{
		}

		void Gen00_Room::NewColumnLoad (Int3 columnId, bool isEditing)
		{
			if (jobMngr->GetJobCount () < 2)
			{
				const bool isAlreadyLoading = SceneCollections::Contains (&columnsLoading, columnId);
				const bool isAlreadyUnloading = SceneCollections::Contains (&columnsUnloading, columnId);

				if (isAlreadyLoading)
				{
					assert (false);
				}
				else if (isAlreadyUnloading)
				{
					// TODO reclaim the unloading column from the unload job queue perhaps instead of waiting to fullyh unload then reload?
					assert (false);
				}
				else
				{
					columnsLoading.push_back (columnId);

					SceneFile::Factory::ColumnLoadJob * job = aah_new SceneFile::Factory::ColumnLoadJob (
						jobMngr,
						room,
						columnId,
						this,
						events,
						isEditing);

					job->AddToQueue (Thread::JB_NORMAL);
				}
			}
		}

		void Gen00_Room::NewColumnUnload (SceneSim::Column * column)
		{
			assert (column != NULL);

			const bool isAlreadyLoading = SceneCollections::Contains (&columnsLoading, column->columnId);
			const bool isAlreadyUnloading = SceneCollections::Contains (&columnsUnloading, column->columnId);

			if (isAlreadyLoading)
			{
				assert (false);
			}
			else if (isAlreadyUnloading)
			{
				assert (false); // this really really should never happen!
			}
			else
			{
				room->RemoveColumn (column);

				columnsUnloading.push_back (column->columnId);

				SceneFile::Factory::ColumnUnloadJob * job = aah_new SceneFile::Factory::ColumnUnloadJob (
					jobMngr,
					column,
					this,
					events);

				job->AddToQueue (Thread::JB_NORMAL);
			}
		}

		void Gen00_Room::NewOnColumnLoadCompleted (SceneSim::Room * room, SceneSim::Column * column)
		{	
			assert (room != NULL);
			assert (column != NULL);
			assert (SceneCollections::Contains (&columnsLoading, column->columnId));

			SceneCollections::Remove (&columnsLoading, column->columnId);

			if (SceneCollections::Contains (&columnsUnloading, column->columnId))
			{
				SceneCollections::Remove (&columnsUnloading, column->columnId);
				delete column;
			}
			else
			{
				room->AddColumn (column);
			}
		}

		void Gen00_Room::NewOnColumnUnloadCompleted (SceneSim::Column * column)
		{	
			assert (column != NULL);
			assert (SceneCollections::Contains (&columnsUnloading, column->columnId));

			SceneCollections::Remove (&columnsUnloading, column->columnId);

			delete column;
		}

		bool Gen00_Room::NewIsColumnRequested (const SceneSim::Room * room, Int3 columnId) const
		{
			assert (room != NULL);

			return SceneCollections::Contains (&columnsLoading, columnId);
		}

		SceneFile::CreateResults Gen00_Room::CreateRoom (SceneSim::WorldInit * worldInit)
		{
			Text roomFolder = world->GetGameFolder () + "data/rooms/" + room->roomId + "/";

			bool isOk = Utils::File_CopyFolder (roomFolder, "data_default/new_room_diorama_00/", false);
			assert (isOk);

			Debug_WriteLine (0, "World created: %s", roomFolder.Raw ());

			Text luaFilename (roomFolder + "/room.lua");
			room->LoadLua (luaFilename);
			Lua::LuaState * lua = room->GetLua ();

			lua->SetString ("room/terrain/primary_generation/random_seed", worldInit->randomSeed);
			lua->SaveText_GlobalTable (luaFilename, "room", Lua::LuaState::WL_NONE);

			LoadDataFromLua ();

			return (isOk) ? SceneFile::CR_SUCCESS : SceneFile::CR_FAIL;
		}

		SceneFile::LoadResults Gen00_Room::LoadRoom ()
		{
			LoadDataFromLua ();

			return SceneFile::LR_SUCCESS;
		}

		void Gen00_Room::LoadDataFromLua ()
		{
			// TODO:
			Lua::LuaState * lua = room->GetLua ();

			newColumnHeight = lua->ReadInt ("room/terrain/column_height");

			Text seedAsText = lua->ReadString ("room/terrain/primary_generation/random_seed");
			randomSeed.s = RakNet::RakString::ToInteger (seedAsText.Raw ());

			landscapeParams.scale = lua->ReadFloat ("room/terrain/primary_generation/landscape/perlin_scale");
			landscapeParams.octavesMax = lua->ReadInt ("room/terrain/primary_generation/landscape/perlin_octave_count");
			landscapeParams.frequencyPerOctave = lua->ReadFloat ("room/terrain/primary_generation/landscape/perlin_frequency_per_octave");
			landscapeParams.amplitudePerOctave = lua->ReadFloat ("room/terrain/primary_generation/landscape/perlin_amplitude_per_octave");

			landscapePerlinScale = lua->ReadFloat ("room/terrain/primary_generation/landscape/perlin_result_scale");
			landscapePerlinTranslate = lua->ReadFloat ("room/terrain/primary_generation/landscape/perlin_result_translate");
			landscapeSolidScale = lua->ReadFloat ("room/terrain/primary_generation/landscape/solid_compare_scale") / (float) (newColumnHeight - 1);
			landscapeSolidTranslate = lua->ReadFloat ("room/terrain/primary_generation/landscape/solid_compare_translate");

			landTypeParams.scale = lua->ReadFloat ("room/terrain/primary_generation/land_type/perlin_scale");
			landTypeParams.octavesMax = lua->ReadInt ("room/terrain/primary_generation/land_type/perlin_octave_count");
			landTypeParams.frequencyPerOctave = lua->ReadFloat ("room/terrain/primary_generation/land_type/perlin_frequency_per_octave");
			landTypeParams.amplitudePerOctave = lua->ReadFloat ("room/terrain/primary_generation/land_type/perlin_amplitude_per_octave");

			landscapeParams.scale = 128.0f;
			landscapeParams.octavesMax = 6;
			landscapeParams.frequencyPerOctave = 2.0f;
			landscapeParams.amplitudePerOctave = 0.5f;

			landscapePerlinScale = 1.2f;
			landscapePerlinTranslate = 0.0f;
			landscapeSolidScale = 0.2f / (float) (newColumnHeight - 1);
			landscapeSolidTranslate = -0.1f;

			landTypeParams.scale = 64.0f;
			landTypeParams.octavesMax = 4;
			landTypeParams.frequencyPerOctave = 2.0f;
			landTypeParams.amplitudePerOctave = 0.5f;
		}

		SceneFile::CreateResults Gen00_Room::CreateColumn (SceneSim::Column * column)
		{
			assert (column != NULL);
			//assert (this->column == NULL);
			//this->column = column;

			//this->column = NULL;
			return SceneFile::CR_SUCCESS;
		}

		SceneFile::LoadResults Gen00_Room::NewAsyncLoadColumn (SceneSim::Column * column, bool isEditing)
		{
			assert (column != NULL);
			//assert (this->column == NULL);
			//this->column = column;

			const int height = room->GetNewColumnHeight ();

			column->Init (height, SceneSim::Column::SF_NONE);

			if (areaMngr->IsColumnLoadable (column->columnId))
			{
				areaMngr->LoadColumn_BaseData (column);

				//this->column = NULL;
				return SceneFile::LR_SUCCESS;
			}

#if defined (DEBUG_SIMPLE_COLUMN)

			const int top = SceneSim::CellsPerColY - 1;
			const int bottom = 0;

			for (int x = 0; x < SceneSim::CellsPerColXz; ++x)
			{
				for (int z = 0; z < SceneSim::CellsPerColXz; ++z)
				{
					for (int y = top; y >= bottom; --y)
					{
						int blockId = 0;
						switch (y)
						{
						case 0: blockId = 1; break;
						case 1: blockId = 78; break;
						default: break;
						}

						SceneSim::Cell * cell = column->GetCells ()->GetCell (x, y, z);
						cell->type = blockId;
						cell->blockData = 0;
						cell->lightSky = 0;
						cell->lightSrc = 0;
					}
				}
			}

			//SceneSim::Cell * cell = column->GetCells ()->GetCell (0, 1, 0);
			//cell->type = 2;

#else

			CalcWeights (column);

			const int top = height - 1;
			const int bottom = 0;
			const int waterLevel = 1;

			for (int x = 0; x < SceneSim::CellsPerColXz; ++x)
			{
				for (int z = 0; z < SceneSim::CellsPerColXz; ++z)
				{
					SceneSim::Cell * cellAbove = NULL;
					int blockedBy = 0;
					int currentSolidStreak = 0;

					for (int y = top; y >= bottom; --y)
					{
						int landTypeIdx = 1; // grass

						const float yy = y / (float) height;
						const float landType = GetLandTypeWeight (x, z);
						if (landType - yy > -0.2f) landTypeIdx = 2; // grass
						if (landType - yy < -0.7f) landTypeIdx = 0; // snowy

						bool isSolid = false;

						int blockId = 0;

						{
							if (y < (height - 1))
							{
								float ground = GetTerrainWeight (x, y, z);
	
								float isSolidWeight = y * landscapeSolidScale + landscapeSolidTranslate;
								ground = ground * landscapePerlinScale + landscapePerlinTranslate;
								isSolid = (ground >= isSolidWeight);
							}

							if (isSolid)
							{	
								++currentSolidStreak;

								int rockBlocksIdx = 0;
								if (currentSolidStreak > 4) rockBlocksIdx = 2;
								else if (currentSolidStreak > 1) rockBlocksIdx = 1;

								blockId = rockBlocks [landTypeIdx][rockBlocksIdx];

								if (cellAbove != NULL && cellAbove->type == 0)
								{
									if (landTypeIdx == 0 && blockedBy == 0)
									{
										//cellAbove->type = 78;
									}
									else 
									{
										int detailRandom = Noise::RandomInt (randomSeed, x * -245 + y * 19 + z * -103);
										if (((detailRandom >> 7) & 0xf) == 0xf)
										{
											if (landTypeIdx == 2)
											{
												cellAbove->type = 81;
											}
											else
											{
												if (((detailRandom >> 17) & 0xf) == 0xf)
												{
													//cellAbove->type = (blockedBy < 4) ? 38 : 40; // red flower / mushroom

												}
												else
												{
													cellAbove->type = 31;
													//cellAbove->type = (blockedBy < 4) ? 37 : 39; // yellow flower / mushroom
												}
											}
										}
									}
								}

								++blockedBy;
							}
							else
							{
								currentSolidStreak = 0;
							}
						}

						if (!isSolid)
						{
							if (y == waterLevel)
							{
								blockId = 9;
							}
							else if (y < waterLevel)
							{
								blockId = 8;
								isSolid = true;
							}
						}

						int light = 0;
						if (!isSolid)
						{
							light = 15 - ((blockedBy + 2) / 3);
							light = (light < 8) ? 8 : light;
						}

						SceneSim::Cell * cell = column->GetCells ()->GetCell (x, y, z);
						cell->type = blockId;
						cell->blockData = 0;
						cell->lightSky = 0;
						cell->lightSrc = 0;

						cellAbove = cell;
					}
				}
			}

			AddTree (column, 2, 0);
			AddTree (column, 2, 0xf4fc2348);
			AddTree (column, 2, 0xa091047f);
			AddTree (column, 2, 0x00342789);
			AddTree (column, 2, 0xfabde83d);

#endif // DEBUG_SIMPLE_COLUMN

			// apply new lighting scheme
			for (int x = 0; x < SceneSim::CellsPerColXz; ++x)
			{
				for (int z = 0; z < SceneSim::CellsPerColXz; ++z)
				{
					Int2 stripId (x, z);
					SceneFile::Lighting::CalcStripSky (column, stripId, room->GetCellBuilder ());
				}
			}

			//column->CalcSkyLightMap (room->GetCellBuilder ());

			// set time to 1 for all newly created builds so that they are saved on quitting.
			RakNet::Time time = 1;
			column->SetLastModifiedTime (time);

			//this->column = NULL;
			return SceneFile::LR_SUCCESS;
		}

		SceneFile::LoadResults Gen00_Room::NewAsyncUnloadColumn (SceneSim::Column * column)
		{
			assert (column != NULL);

			areaMngr->SaveColumn (column);
			return SceneFile::LR_SUCCESS;
		}

		void Gen00_Room::UnloadColumn (SceneSim::Column * column)
		{
			assert (false);

			if (!SceneCollections::Contains (&columnsLoading, column->columnId))
			{
				//std::cout << "column UNLOADED (" << 
				//	column->columnId.x << ", " <<
				//	column->columnId.y << ", " <<
				//	column->columnId.z << ")" << std::endl;

				room->RemoveColumn (column);
				delete column;
			}
			else
			{
				columnsUnloading.push_back (column->columnId);

				assert (false);
				//waitingToUnload->Add (column);
			}
		}

		void Gen00_Room::CalcWeights (SceneSim::Column * column)
		{
			const int startX = column->columnId.x * SceneSim::CellsPerColXz;
			const int startY = 0;// column->columnId.y * SceneSim::CellsPerColY;
			const int startZ = column->columnId.z * SceneSim::CellsPerColXz;

			for (int x = 0; x < smoothX; ++x)
			{
				for (int z = 0; z < smoothZ; ++z)
				{
					const float xx = (float) (startX + x * smoothRes);
					const float zz = (float) (startZ + z * smoothRes);

					landTypeWeights [x][z] = Noise::Perlin2d (randomSeed, xx + 57654.0f, zz - 2554.0f, &landTypeParams);

					for (int y = 0; y < smoothY; ++y)
					{
						const float yy = (float) (startY + y * smoothRes);
						terrainWeights [x][y][z] = Noise::Perlin3d (randomSeed, xx - 432.0f, yy + 3432.0f, zz - 546547.0f, &landscapeParams);
					}
				}
			}
		}

		float Gen00_Room::GetLandTypeWeight (int x, int z)
		{
			int xs = x / smoothRes;
			int zs = z / smoothRes;
			float xw = (x % smoothRes) / (float) smoothRes;
			float zw = (z % smoothRes) / (float) smoothRes;

			float result = SmoothWeight2d (
				landTypeWeights [xs][zs],
				landTypeWeights [xs + 1][zs],
				landTypeWeights [xs][zs + 1],
				landTypeWeights [xs + 1][zs + 1],
				xw, zw);

			return result;
		}

		float Gen00_Room::GetTerrainWeight (int x, int y, int z)
		{
			int xs = x / smoothRes;
			int ys = y / smoothRes;
			int zs = z / smoothRes;
			float xw = (x % smoothRes) / (float) smoothRes;
			float yw = (y % smoothRes) / (float) smoothRes;
			float zw = (z % smoothRes) / (float) smoothRes;

			float plane1 = SmoothWeight2d (
				terrainWeights [xs][ys][zs],
				terrainWeights [xs + 1][ys][zs],
				terrainWeights [xs][ys][zs + 1],
				terrainWeights [xs + 1][ys][zs + 1],
				xw, zw);

			float plane2 = SmoothWeight2d (
				terrainWeights [xs][ys + 1][zs],
				terrainWeights [xs + 1][ys + 1][zs],
				terrainWeights [xs][ys + 1][zs + 1],
				terrainWeights [xs + 1][ys + 1][zs + 1],
				xw, zw);

			float result = SmoothWeight (plane1, plane2, yw);
			return result;
		}

		float Gen00_Room::SmoothWeight (float a, float b, float weight) const
		{
			return  b * weight + a * (1.0f - weight);
		}

		float Gen00_Room::SmoothWeight2d (float a1, float b1, float a2, float b2, float weight1, float weight2) const
		{
			float s1 = SmoothWeight (a1, b1, weight1);
			float s2 = SmoothWeight (a2, b2, weight1);
			return SmoothWeight (s1, s2, weight2);
		}

		void Gen00_Room::AddTree (SceneSim::Column * column, int waterLevel, unsigned int mask)
		{
			const int height = column->GetSize ().y;

			int size = 2;			
			unsigned int treeRand = Noise::RandomInt (randomSeed, column->columnId.x * 57 + column->columnId.z * 7);
			treeRand ^= mask;

			if (((treeRand >> 7) % 4) != 0)
			{
				unsigned int treeX = (treeRand >> 4) % (16 - size - size - 1);
				unsigned int treeZ = (treeRand >> 11) % (16 - size - size - 1);

				treeX += size + 1;
				treeZ += size + 1;
				
				// add tree
				Int3 trunkTop (treeX, 0, treeZ);
				bool treeFound = false;
				bool isCactus = false;
				for (int y = height - 10; y > (waterLevel + 1); --y)
				{
					SceneSim::Cell * cell = column->GetCells ()->GetCell (treeX, y, treeZ);
					if (cell->type == 2 || cell->type == 12)
					{
						isCactus = (cell->type == 12);

						// add tree!
						int trunkHeight = 6;
						
						trunkTop.y = y + trunkHeight;
						for (int trunkY = trunkTop.y; trunkY > y; --trunkY)
						{
							SceneSim::Cell * cell = column->GetCells ()->GetCell (treeX, trunkY, treeZ);
							cell->type = (isCactus) ? 81 : 17;
							cell->blockData = 0;
							//cell->lightSky = 15;
							cell->lightSrc = 0;
						}
						treeFound = true;
						break;
					}
				}

				if (treeFound && !isCactus)
				{
					float sizeSqr = (size + 0.5f) * (size + 0.5f);
					for (int x = trunkTop.x - size; x <= trunkTop.x + size; ++x)
					{
						for (int y = trunkTop.y - size; y <= trunkTop.y + size; ++y)
						{
							for (int z = trunkTop.z - size; z <= trunkTop.z + size; ++z)
							{
								SceneSim::Cell * cell = column->GetCells ()->GetCell (x, y, z);
								if (cell->type == 0)
								{
									cell->type = 18;
								}
							}
						}
					}
				}
			}
		}
	}
}
