#include "stdafx.h"

#include "Lua/Lua_LuaState.h"
#include "Thread/Thread_JobMngr.h"

#include "../Noise/Noise_Perlin.h"
#include "../Noise/Noise_Random.h"
#include "../SceneCollections/SceneCollections_Utils.h"
#include "../SceneFile/Area_AreaMngr.h"
#include "../SceneFile/Factory_ColumnLoadJob.h"
#include "../SceneFile/Factory_ColumnUnloadJob.h"
#include "../SceneFile/SceneFile_Lighting.h"
#include "../SceneSim/SceneSim_Cell.h"
#include "../SceneSim/SceneSim_Column.h"
#include "../SceneSim/SceneSim_ColumnCells.h"
#include "../SceneSim/SceneSim_Room.h"
#include "../SceneSim/SceneSim_World.h"
#include "../SceneSim/SceneSim_WorldInit.h"

#include "Ld00_Room.h"

//#define DEBUG_SIMPLE_COLUMN

namespace SceneGen
{
	namespace Ld_00
	{
		const int rockBlocks [][3] =
		{
			{128, 3, 57/*79*/}, // snow
			{2, 3, 1}, // grass
			{12, 24, 4}, // sand
		};

		Ld_00_Room::Ld_00_Room (SceneSim::Room * room, Thread::JobMngr * const jobMngr, SceneSim::World * world, SceneHelper::SimEvents * events)
			: OperationsRoom (room, jobMngr, world, events),
			randomSeed (0)
		{
		}

		Ld_00_Room::~Ld_00_Room ()
		{
		}

		void Ld_00_Room::NewColumnLoad (Int3 columnId, bool isEditing)
		{
			if (jobMngr->GetJobCount () < 2)
			{
				const bool isAlreadyLoading = SceneCollections::Contains (&columnsLoading, columnId);
				const bool isAlreadyUnloading = SceneCollections::Contains (&columnsUnloading, columnId);

				if (isAlreadyLoading)
				{
					assert (false);
				}
				else if (isAlreadyUnloading)
				{
					// TODO reclaim the unloading column from the unload job queue perhaps instead of waiting to fullyh unload then reload?
					assert (false);
				}
				else
				{
					columnsLoading.push_back (columnId);

					SceneFile::Factory::ColumnLoadJob * job = aah_new SceneFile::Factory::ColumnLoadJob (
						jobMngr,
						room,
						columnId,
						this,
						events,
						isEditing);

					job->AddToQueue (Thread::JB_NORMAL);
				}
			}
		}

		void Ld_00_Room::NewColumnUnload (SceneSim::Column * column)
		{
			assert (column != NULL);

			const bool isAlreadyLoading = SceneCollections::Contains (&columnsLoading, column->columnId);
			const bool isAlreadyUnloading = SceneCollections::Contains (&columnsUnloading, column->columnId);

			if (isAlreadyLoading)
			{
				assert (false);
			}
			else if (isAlreadyUnloading)
			{
				assert (false); // this really really should never happen!
			}
			else
			{
				room->RemoveColumn (column);

				columnsUnloading.push_back (column->columnId);

				SceneFile::Factory::ColumnUnloadJob * job = aah_new SceneFile::Factory::ColumnUnloadJob (
					jobMngr,
					column,
					this,
					events);

				job->AddToQueue (Thread::JB_NORMAL);
			}
		}

		void Ld_00_Room::NewOnColumnLoadCompleted (SceneSim::Room * room, SceneSim::Column * column)
		{	
			assert (room != NULL);
			assert (column != NULL);
			assert (SceneCollections::Contains (&columnsLoading, column->columnId));

			SceneCollections::Remove (&columnsLoading, column->columnId);

			if (SceneCollections::Contains (&columnsUnloading, column->columnId))
			{
				SceneCollections::Remove (&columnsUnloading, column->columnId);
				delete column;
			}
			else
			{
				room->AddColumn (column);
			}
		}

		void Ld_00_Room::NewOnColumnUnloadCompleted (SceneSim::Column * column)
		{	
			assert (column != NULL);
			assert (SceneCollections::Contains (&columnsUnloading, column->columnId));

			SceneCollections::Remove (&columnsUnloading, column->columnId);

			delete column;
		}

		bool Ld_00_Room::NewIsColumnRequested (const SceneSim::Room * room, Int3 columnId) const
		{
			assert (room != NULL);

			return SceneCollections::Contains (&columnsLoading, columnId);
		}

		SceneFile::CreateResults Ld_00_Room::CreateRoom (SceneSim::WorldInit * worldInit)
		{
			Text roomFolder = world->GetGameFolder () + "data/rooms/" + room->roomId + "/";

			bool isOk = Utils::File_CopyFolder (roomFolder, "data_default/new_room_diorama_00/", false);
			assert (isOk);

			Debug_WriteLine (0, "World created: %s", roomFolder.Raw ());

			Text luaFilename (roomFolder + "/room.lua");
			room->LoadLua (luaFilename);
			Lua::LuaState * lua = room->GetLua ();

			lua->SetString ("room/terrain/primary_generation/random_seed", worldInit->randomSeed);
			lua->SaveText_GlobalTable (luaFilename, "room", Lua::LuaState::WL_NONE);

			LoadDataFromLua ();

			return (isOk) ? SceneFile::CR_SUCCESS : SceneFile::CR_FAIL;
		}

		SceneFile::LoadResults Ld_00_Room::LoadRoom ()
		{
			LoadDataFromLua ();

			return SceneFile::LR_SUCCESS;
		}

		void Ld_00_Room::LoadDataFromLua ()
		{
			// TODO:
			Lua::LuaState * lua = room->GetLua ();

			newColumnHeight = lua->ReadInt ("room/terrain/column_height");

			Text seedAsText = lua->ReadString ("room/terrain/primary_generation/random_seed");
			randomSeed.s = RakNet::RakString::ToInteger (seedAsText.Raw ());

			landscapeParams.scale = lua->ReadFloat ("room/terrain/primary_generation/landscape/perlin_scale");
			landscapeParams.octavesMax = lua->ReadInt ("room/terrain/primary_generation/landscape/perlin_octave_count");
			landscapeParams.frequencyPerOctave = lua->ReadFloat ("room/terrain/primary_generation/landscape/perlin_frequency_per_octave");
			landscapeParams.amplitudePerOctave = lua->ReadFloat ("room/terrain/primary_generation/landscape/perlin_amplitude_per_octave");

			landscapePerlinScale = lua->ReadFloat ("room/terrain/primary_generation/landscape/perlin_result_scale");
			landscapePerlinTranslate = lua->ReadFloat ("room/terrain/primary_generation/landscape/perlin_result_translate");
			landscapeSolidScale = lua->ReadFloat ("room/terrain/primary_generation/landscape/solid_compare_scale") / (float) (newColumnHeight - 1);
			landscapeSolidTranslate = lua->ReadFloat ("room/terrain/primary_generation/landscape/solid_compare_translate");

			landTypeParams.scale = lua->ReadFloat ("room/terrain/primary_generation/land_type/perlin_scale");
			landTypeParams.octavesMax = lua->ReadInt ("room/terrain/primary_generation/land_type/perlin_octave_count");
			landTypeParams.frequencyPerOctave = lua->ReadFloat ("room/terrain/primary_generation/land_type/perlin_frequency_per_octave");
			landTypeParams.amplitudePerOctave = lua->ReadFloat ("room/terrain/primary_generation/land_type/perlin_amplitude_per_octave");

			landscapeParams.scale = 128.0f;
			landscapeParams.octavesMax = 6;
			landscapeParams.frequencyPerOctave = 2.0f;
			landscapeParams.amplitudePerOctave = 0.5f;

			landscapePerlinScale = 1.2f;
			landscapePerlinTranslate = 0.0f;
			landscapeSolidScale = 0.2f / (float) (newColumnHeight - 1);
			landscapeSolidTranslate = -0.1f;

			landTypeParams.scale = 64.0f;
			landTypeParams.octavesMax = 4;
			landTypeParams.frequencyPerOctave = 2.0f;
			landTypeParams.amplitudePerOctave = 0.5f;
		}

		SceneFile::CreateResults Ld_00_Room::CreateColumn (SceneSim::Column * column)
		{
			assert (column != NULL);
			//assert (this->column == NULL);
			//this->column = column;

			//this->column = NULL;
			return SceneFile::CR_SUCCESS;
		}

		SceneFile::LoadResults Ld_00_Room::NewAsyncLoadColumn (SceneSim::Column * column, bool isEditing)
		{
			assert (column != NULL);
			//assert (this->column == NULL);
			//this->column = column;

			const int height = room->GetNewColumnHeight ();

			column->Init (height, SceneSim::Column::SF_NONE);

			const int top = SceneSim::CellsPerColY - 1;
			const int bottom = 0;

			if (areaMngr->IsColumnLoadable (column->columnId))
			{
				areaMngr->LoadColumn_BaseData (column);

				if (!isEditing)
				{
					// remove all scaffolfding
					for (int x = 0; x < SceneSim::CellsPerColXz; ++x)
					{
						for (int z = 0; z < SceneSim::CellsPerColXz; ++z)
						{
							for (int y = top; y >= bottom; --y)
							{
								SceneSim::Cell * cell = column->GetCells ()->GetCell (x, y, z);
								if (cell->type == 1)
								{
									cell->type = 0;
								}
							}
						}
					}
				}

				// apply new lighting scheme
				for (int x = 0; x < SceneSim::CellsPerColXz; ++x)
				{
					for (int z = 0; z < SceneSim::CellsPerColXz; ++z)
					{
						Int2 stripId (x, z);
						SceneFile::Lighting::CalcStripSky (column, stripId, room->GetCellBuilder ());
					}
				}

				//this->column = NULL;
				return SceneFile::LR_SUCCESS;
			}

			for (int x = 0; x < SceneSim::CellsPerColXz; ++x)
			{
				for (int z = 0; z < SceneSim::CellsPerColXz; ++z)
				{
					for (int y = top; y >= bottom; --y)
					{
						int blockId = 0;
						if (y == 0)
						{
							blockId = 1;
						}
						else if ((y % 8) == 0)
						{
							blockId = (x == 0 || z == 0) ? 1 : 0;
						}

						SceneSim::Cell * cell = column->GetCells ()->GetCell (x, y, z);
						cell->type = blockId;
						cell->blockData = 0;
						cell->lightSky = 0;
						cell->lightSrc = 0;
					}
				}
			}

			//SceneSim::Cell * cell = column->GetCells ()->GetCell (0, 1, 0);
			//cell->type = 2;

			// apply new lighting scheme
			for (int x = 0; x < SceneSim::CellsPerColXz; ++x)
			{
				for (int z = 0; z < SceneSim::CellsPerColXz; ++z)
				{
					Int2 stripId (x, z);
					SceneFile::Lighting::CalcStripSky (column, stripId, room->GetCellBuilder ());
				}
			}

			//column->CalcSkyLightMap (room->GetCellBuilder ());

			// set time to 1 for all newly created builds so that they are saved on quitting.
			RakNet::Time time = 1;
			column->SetLastModifiedTime (time);

			//this->column = NULL;
			return SceneFile::LR_SUCCESS;
		}

		SceneFile::LoadResults Ld_00_Room::NewAsyncUnloadColumn (SceneSim::Column * column)
		{
			assert (column != NULL);

			areaMngr->SaveColumn (column);
			return SceneFile::LR_SUCCESS;
		}

		void Ld_00_Room::UnloadColumn (SceneSim::Column * column)
		{
			assert (false);

			if (!SceneCollections::Contains (&columnsLoading, column->columnId))
			{
				//std::cout << "column UNLOADED (" << 
				//	column->columnId.x << ", " <<
				//	column->columnId.y << ", " <<
				//	column->columnId.z << ")" << std::endl;

				room->RemoveColumn (column);
				delete column;
			}
			else
			{
				columnsUnloading.push_back (column->columnId);

				assert (false);
				//waitingToUnload->Add (column);
			}
		}
	}
}
