#include "stdafx.h"

#include <algorithm>
#include <iostream>
#include <fstream>

#include "File_PakFileBase.h"
#include "File_PakFileRecord.h"

namespace File
{
	unsigned char PakFileBase::padBuffer [blockSize];

	PakFileBase::PakFileBase (TextRef filename, int recordCount)
		: filename (filename),
		records (NULL)
	{
		std::ifstream file;
		file.open (filename.Raw (), std::ios_base::binary);

		const int begin = file.tellg ();
		if (begin != -1)
		{
			file.read ((char *) &header, sizeof (PakFileHeader));
			records = aah_new PakFileRecord [header.recordCount];
			file.read ((char *) records, sizeof (PakFileRecord) * header.recordCount);
			file.close ();

			sortedRecordIds = aah_new int [header.recordCount];

			// ludum dare - bodge cos of switching to queryperformancecounter which resets to zero every pc reset!
			for (int idx = 0; idx < header.recordCount; ++idx)
			{
				records [idx].lastModified = RakNet::Time (0);
			}
		}
		else
		{
			InitNewHeader (recordCount);
			SaveHeader (true);
		}

		for (int idx = 0; idx < header.recordCount; ++idx)
		{
			sortedRecordIds [idx] = idx;
		}

		memset (&padBuffer [0], 0, sizeof (padBuffer));
	}

	PakFileBase::~PakFileBase ()
	{
		delete [] sortedRecordIds;
		delete [] records;
	}

	bool PakFileBase::DoesRecordExist (int recordIdx) const
	{
		const PakFileRecord * record = &records [recordIdx];
		
		return (record->blockOffset != -1);
	}

	int PakFileBase::LoadRecord (int recordIdx, RakNet::Time * lastModified, unsigned char * out, int maxOut)
	{
		assert (recordIdx >= 0 && recordIdx < header.recordCount);
		assert (lastModified != NULL);
		assert (out != NULL);
		assert (maxOut > 0);
		assert (DoesRecordExist (recordIdx));

		PakFileRecord * record = &records [recordIdx];
		assert (maxOut >= record->fileLength);

		std::fstream file;
		file.open (filename.Raw (), std::ios_base::in | std::ios_base::binary);

		file.seekg (header.firstBlockOffset + record->blockOffset * blockSize, std::ios_base::beg);
		file.read ((char *) out, record->fileLength);

		file.close ();

		//Debug_PrintChecksum (recordIdx, out, record->fileLength);
		
		*lastModified = record->lastModified;

		return record->fileLength;
	}

	void PakFileBase::SaveRecord (int recordIdx, const RakNet::Time * lastModified, unsigned char * in, int size)
	{
		assert (recordIdx >= 0 && recordIdx < header.recordCount);
		assert (lastModified != NULL);
		assert (in != NULL);

		PakFileRecord * record = &records [recordIdx];

		if (*lastModified > record->lastModified)
		{
			//Debug_PrintChecksum (recordIdx, in, size);

			const int oldBlockSize = record->blockLength;
			const int oldBlockOffset = record->blockOffset;
			const int newBlockSize = (size + blockSize - 1) / blockSize;
			
			if (oldBlockSize != newBlockSize)
			{
				// find a better place to put the data (if there is one) by releasing the old blocks then searching
				record->blockOffset = -1;
				record->blockLength = -1;
				record->fileLength = -1;

				record->blockOffset = FindFirstFreeBlocks (newBlockSize);
				record->blockLength = newBlockSize;
			}

			record->fileLength = size;
			record->lastModified = *lastModified;

			std::fstream file;
			file.open (filename.Raw (), std::ios_base::in | std::ios_base::out | std::ios_base::binary);

			// save header
			file.write ((char *) &header, sizeof (PakFileHeader));
			file.write ((char *) records, sizeof (PakFileRecord) * header.recordCount);

			// save column data
			file.seekp (header.firstBlockOffset + record->blockOffset * blockSize, std::ios_base::beg);
			file.write ((char *) in, size);
			PadToNextBlock (&file);

			file.close ();
		}
	}

	void PakFileBase::InitNewHeader (int recordCount)
	{
		strncpy (header.tag, "pakf", 4);
		header.versionNumber = 1;
		header.recordCount = recordCount;
		header.firstBlockOffset = GetMinBlockCount (sizeof (PakFileHeader) + sizeof (PakFileRecord) * recordCount) * blockSize;

		records = aah_new PakFileRecord [recordCount];
		sortedRecordIds = aah_new int [recordCount];
	}

	void PakFileBase::SaveHeader (bool addPadding)
	{
		std::fstream file;
		file.open (filename.Raw (), std::ios_base::out | std::ios_base::binary);

		file.write ((char *) &header, sizeof (PakFileHeader));
		file.write ((char *) records, sizeof (PakFileRecord) * header.recordCount);

		if (addPadding)
		{
			PadToNextBlock (&file);
		}

		file.close ();
	}

	void PakFileBase::PadToNextBlock (std::fstream * file)
	{
		const int pos = file->tellp ();
		const int padWith = blockSize - ((pos + blockSize - 1) % blockSize) - 1;

		file->write ((char *) &padBuffer [0], padWith);
	}

	int PakFileBase::GetMinBlockCount (int size) const
	{
		const int result = (size + blockSize - 1) / blockSize;
		return result;
	}

	struct CompareFunction
	{
		CompareFunction (PakFileRecord * records) : records (records) {}

		bool operator () (int i,int j) 
		{
			return (records [i].blockOffset < records [j].blockOffset);
		}
		
		PakFileRecord * records;
	};

	int PakFileBase::FindFirstFreeBlocks (int blockCount) const
	{
		std::sort (sortedRecordIds, sortedRecordIds + header.recordCount, CompareFunction (records));

		int idx = 0;
		while (true)
		{
			if (idx == header.recordCount)
			{
				return 0;
			}
			PakFileRecord * record = &records [sortedRecordIds [idx]];
			if (record->blockLength != -1)
			{
				break;
			}
			++idx;
		}

		while (true)
		{
			PakFileRecord * record = &records [sortedRecordIds [idx]];
			if (idx == header.recordCount - 1)
			{
				return (record->blockOffset + record->blockLength);
			}
			PakFileRecord * recordNext = &records [sortedRecordIds [idx + 1]];
			if (recordNext->blockOffset - record->blockOffset - record->blockLength >= blockCount)
			{
				return (record->blockOffset + record->blockLength);
			}

			++idx;
		}
	}

	unsigned int PakFileBase::Debug_GetChecksum (int recordIdx, unsigned char * data, int length) const
	{
		unsigned int checksum = 0;
		for (int idx = 0; idx < length; ++idx)
		{
			checksum = checksum + data [idx];
		}

		return checksum;
		
		//printf ("CHECKSUM area recordIdx (%d) result (%d)\n",
		//	recordIdx,
		//	checksum);
	}
}
