#include "stdafx.h"

#include "Thread_JobMngr.h"

//#define THREAD_PREVENT_MULTITHREADING
//#define THREAD_JOBMNGR

#if defined (THREAD_JOBMNGR)
#	define THREAD_JOBMNGR_DEBUG(_txt)	std::cout << _txt << std::endl
#else
#	define THREAD_JOBMNGR_DEBUG(_txt)	while (false) {}
#endif

namespace Thread
{
	JobMngr::JobMngr (const char * debugName, bool isDebugging)
		: debugName (strdup (debugName)),
		isDebugging (isDebugging),
		state (S_NONE),
		jobsPendingAndCompletedTotal (0)
	{

#if defined (THREAD_PREVENT_MULTITHREADING)

		isDebugging = this->isDebugging = true;

#endif

		THREAD_JOBMNGR_DEBUG ("JobMngr (" << debugName << " ) created by thread " << GetCurrentThreadId ());
		if (!isDebugging)
		{
			thread = SDL_CreateThread (RunThreadStatic, this);
		}
	}

	JobMngr::~JobMngr ()
	{
		assert (state == S_QUIT);

		if (!isDebugging)
		{
			int result = -1;
			SDL_WaitThread (thread, &result);
			assert (result == 0);
		}

		delete [] debugName;
	}

	int JobMngr::GetJobCount () const
	{
		return jobsPendingAndCompletedTotal;
	}

	void JobMngr::DeleteJob (Job * job)
	{
		assert (job);

		jobsAwaitingDelete.Remove (job);

		delete job;
	}

	void JobMngr::MainThread_Update ()
	{
		if (isDebugging)
		{
			while (true)
			{
				Job * job = jobsPending.PopFront ();

				if (job == NULL)
				{
					break;
				}
				else
				{
					job->state = JS_WORKING;
					const JOBSTATES result = job->JobDoWork ();
					assert (result == JS_WORKDONE);
					job->state = result;

					jobsCompleted.Push (job);
				}
			}
		}

		while (true)
		{
			Job * job = jobsCompleted.Pop ();
			if (job == NULL)
			{
				break;
			}

			assert (job->state == JS_WORKDONE);
			const JOBSTATES result = job->JobOnCompleted ();
			assert (result == JS_COMPLETED);
			job->state = result;

			if (job->deleteSelfOnCompletion)
			{
				delete job;
			}
			else
			{
				jobsAwaitingDelete.PushBack (job);
			}

			--jobsPendingAndCompletedTotal;
		}
	}

	void JobMngr::AddJob (Job * job, JOBPRIORITIES priority)
	{
		assert (job != NULL);
		assert (job->state == JS_NONE);

		//THREAD_JOBMNGR_DEBUG ("Job added (" << debugName);

		++jobsPendingAndCompletedTotal;

		job->state = JS_QUEUED;

		switch (priority)
		{
		case JB_HIGH: jobsPending.PushFront (job); break;
		case JB_NORMAL: jobsPending.PushBack (job); break;
		default: assert (false);
		}
	}

	void JobMngr::FinishCurrentJobAndDeleteAll ()
	{
		if (!isDebugging)
		{
			assert (state == S_RUNNING);

			state = S_QUITTING;

			while (state == S_QUITTING)
			{
				Sleep (10);
			}
			assert (state == S_QUIT);
		}
		else
		{
			state = S_QUIT;
		}

		// clear all queues of incomplete jobs
		while (true)
		{
			Job * job = jobsPending.PopFront ();
			if (job == NULL)
			{
				break;
			}
			delete job;
			--jobsPendingAndCompletedTotal;
		}

		while (true)
		{
			Job * job = jobsCompleted.Pop ();
			if (job == NULL)
			{
				break;
			}
			delete job;
			--jobsPendingAndCompletedTotal;
		}

		while (true)
		{
			Job * job = jobsAwaitingDelete.PopFront ();
			if (job == NULL)
			{
				break;
			}
			delete job;
		}

		assert (jobsPendingAndCompletedTotal == 0);
	}

	int JobMngr::RunThreadStatic (void * data)
	{
		assert (data != NULL);

		JobMngr * jobMngr = (JobMngr *) data;
		return jobMngr->RunThread ();
	}

	int JobMngr::RunThread ()
	{
		state = S_RUNNING;

		THREAD_JOBMNGR_DEBUG ("JobMngr (" << debugName << " ) is on thread " << GetCurrentThreadId ());

		while (state != S_QUITTING)
		{
			Job * job = jobsPending.PopFront ();

			if (job == NULL)
			{
				Sleep (10);
			}
			else
			{
				job->state = JS_WORKING;
				JOBSTATES result = job->JobDoWork ();
				assert (result == JS_WORKDONE);
				job->state = result;

				jobsCompleted.Push (job);
			}
		}
		
		state = S_QUIT;

		return 0;
	}
}
