#include "stdafx.h"

#include <algorithm>

#include "SDL/SDL-1.2.14 sdk/include/SDL.h" // main SDL header

#include "Utils_Text.h"

#if defined (TEXT_DEBUG_SAFE_TYPE)

namespace RakNet
{
	namespace RakString
	{
		int ToInteger (const char * str)
		{
			// this is a hash function - not straight atoi
			
			unsigned long hash = 5381;
			int c;

			while (c = *str++)
			{
				hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
			}

			return (int) hash;
		}
	}
}

#endif

namespace Utils
{
	TextString TextString::NonVariadic (const char * str)
	{
		Text text;
		text.RawSet (str);
		return text;
	}

	TextString TextString::Format (const char * fmt, ...)
	{
		va_list args;
		va_start (args, fmt);
		Text text = Format (fmt, args);
		va_end (args);

		return text;
	}

	TextString TextString::Format (const char * fmt, va_list args)
	{
		static const int bufferMax = 512;
		char buffer [bufferMax];

		vsprintf (buffer, fmt, args);
		assert (strlen (buffer) < bufferMax - 1);
		TextString text;
		text.RawSet (buffer);

		return text;
	}

	bool TextString::ValidateChar (VALIDATETYPES type, char c)
	{
		bool isAllowed = false;
		switch (type)
		{
		case VT_NUMBER:
			if (c >= SDLK_0 && c <= SDLK_9)
			{
				isAllowed = true;
			}
			break;
		
		case VT_IP:
			if ((c >= SDLK_0 && c <= SDLK_9) || c == SDLK_PERIOD)
			{
				isAllowed = true;
			}
			break;

		case VT_ALPHANUMERIC:
			if ((c >= SDLK_0 && c <= SDLK_9) || 
				(c >= SDLK_a && c <= SDLK_z) ||
				(c >= 'A' && c <= 'Z'))
			{
				isAllowed = true;
			}
			break;

		case VT_FILENAME:
			if ((c >= SDLK_0 && c <= SDLK_9) || 
				(c >= SDLK_a && c <= SDLK_z) ||
				(c >= 'A' && c <= 'Z') ||
				c == SDLK_UNDERSCORE ||
				c == SDLK_SPACE)
			{
				isAllowed = true;
			}
			break;

		case VT_FILEPATH:
			if ((c >= SDLK_0 && c <= SDLK_9) || 
				(c >= SDLK_a && c <= SDLK_z) ||
				(c >= 'A' && c <= 'Z') ||
				c == SDLK_PERIOD ||
				c == SDLK_SLASH ||
				c == SDLK_UNDERSCORE ||
				c == SDLK_SPACE)
			{
				isAllowed = true;
			}
			break;

		case VT_TEXT:
			if ((c >= SDLK_0 && c <= SDLK_9) || 
				(c >= SDLK_a && c <= SDLK_z) ||
				(c >= 'A' && c <= 'Z') ||
				c == SDLK_UNDERSCORE)
			{
				isAllowed = true;
			}
			break;

		default:
			assert (false);
		}
		return isAllowed;
	}

	TextString TextString::Validate (VALIDATETYPES type, const char * in, char replaceWith)
	{
		assert (in != NULL);

		static char buffer [256];
		int idx = 0;
		for (; in [idx] != '\0'; ++idx)
		{
			buffer [idx] = (ValidateChar (type, in [idx])) ? in [idx] : replaceWith;
		}
		buffer [idx] ='\0';
		return TextString (buffer);
	}


#if defined (TEXT_DEBUG_SAFE_TYPE)

	TextString::TextString ()
	{
	}

	TextString::TextString (const char * rhs)
		: std::string (rhs)
	{
	}

	TextString::~TextString ()
	{
	}

	TextString::operator const char * () const
	{
		return Raw ();
	}

	void TextString::Clear ()
	{
		assert (false);
	}

	void TextString::Set (const char * fmt, ...)
	{
		assert (false);
	}

	void TextString::Set (const TextString &)
	{
		assert (false);
	}

	void TextString::RawSet (const char * str)
	{
		*this = str;
	}

	TextString TextString::ReplaceStr(const char * find, const char * replace) const
	{
		assert (strlen (find) == 1 && strlen (replace) == 1);

		TextString str = *this;

		std::replace (str.begin (), str.end(), *find, *replace);

		return str;
	}

	bool TextString::operator < (const TextString & rhs) const
	{
		return ((*(std::string *) this) < (const std::string &) rhs);
	}

	bool TextString::operator <= (const TextString & rhs) const
	{
		return ((*(std::string *) this) <= (const std::string &) rhs);
	}

	bool TextString::operator > (const TextString & rhs) const
	{
		return ((*(std::string *) this) > (const std::string &) rhs);
	}

	bool TextString::operator >= (const TextString & rhs) const
	{
		return ((*(std::string *) this) >= (const std::string &) rhs);
	}

	void TextString::Truncate (int length)
	{
		assert (length >= 0);

		if (Length () > (length - 1))
		{
			erase (length);
		}
	}

	int TextString::Length () const
	{
		return size ();
	}

	TextString TextString::SubStr (int pos, int len) const
	{
		std::string str (substr (pos, len));
		return *((TextString*) &str);
	}

	const char * TextString::Raw () const
	{
		return c_str ();
	}

	TextString TextString::operator + (const char * rhs) const
	{
		TextString text (*this);
		text += rhs;
		return text;
	}

	TextString TextString::operator + (const TextString & rhs) const
	{
		TextString text (*this);
		text += rhs;
		return text;
	}

	TextString & TextString::operator += (char rhs)
	{
		this->std::string::operator += (rhs);
		return *this;
	}

	TextString & TextString::operator += (const TextString & rhs)
	{
		this->std::string::operator += (rhs.Raw ());
		return *this;
	}

	bool TextString::operator == (const char * rhs) const
	{
		return ((*(std::string *) this) == rhs);
	}
	
	bool TextString::operator == (const TextString & rhs) const
	{
		return ((*(std::string *) this) == (const std::string &) rhs);
	}

	bool TextString::operator != (const char * rhs) const
	{
		return ((*(std::string *) this) != rhs);
	}

	bool TextString::operator != (const TextString & rhs) const
	{
		return ((*(std::string *) this) != (const std::string &) rhs);
	}

#else

	TextString::TextString ()
	{
	}

	TextString::TextString (const char * rhs)
		: RakNet::RakString (rhs)
	{
	}

	TextString::~TextString ()
	{
	}

	TextString::operator const char * () const
	{
		return C_String ();
	}

	void TextString::Clear ()
	{
		RakNet::RakString::Clear ();
	}

	void TextString::Set (const char * fmt, ...)
	{
		va_list args;
		va_start (args, fmt);
		Text other = Format (fmt, args);
		va_end (args);
		RakNet::RakString::Set (other.C_String ());
	}

	void TextString::Set (const TextString &)
	{
		assert (false);
	}

	void TextString::RawSet (const char * str)
	{
		(*(RakNet::RakString *) this) = str;
	}

	bool TextString::operator < (const TextString & rhs) const
	{
		return ((*(RakNet::RakString *) this) < (const RakNet::RakString &) rhs);
	}

	bool TextString::operator <= (const TextString & rhs) const
	{
		return ((*(RakNet::RakString *) this) <= (const RakNet::RakString &) rhs);
	}

	bool TextString::operator > (const TextString & rhs) const
	{
		return ((*(RakNet::RakString *) this) > (const RakNet::RakString &) rhs);
	}

	bool TextString::operator >= (const TextString & rhs) const
	{
		return ((*(RakNet::RakString *) this) >= (const RakNet::RakString &) rhs);
	}


	void TextString::Truncate (int length)
	{
		RakNet::RakString::Truncate (length);
	}

	int TextString::Length () const
	{
		return RakNet::RakString::GetLength ();
	}

	TextString TextString::SubStr (int pos, int len) const
	{
		TextString text = ((RakNet::RakString *) this)->SubStr (pos, len);
		return text;
	}

	TextString TextString::ReplaceStr (const char * find, const char * replace) const
	{
		assert (find != NULL);
		assert (replace != NULL);

		static const int bufferMax = 512;
		char buffer [bufferMax];

		buffer [0] = '\0';

		const char * c = Raw ();
		for (const char * c = Raw (); c != NULL; )
		{
			const char * oc = c;
			c = strstr (c, find);
			if (c == NULL)
			{
				strncat (buffer, oc, Raw () + strlen (Raw ()) - oc);
				break;
			}
			strncat (buffer, oc, c - oc);
			strcat (buffer, replace);

			c += strlen (find);
		}

		return TextString (buffer);
	}

	const char * TextString::Raw () const
	{
		return C_String ();
	}

	TextString TextString::operator + (const char * rhs) const
	{
		TextString text (*this);
		text += rhs;
		return text;
	}

	TextString TextString::operator + (const TextString & rhs) const
	{
		TextString text (*this);
		text += rhs;
		return text;
	}

	TextString & TextString::operator += (char rhs)
	{
		this->RakNet::RakString::operator += (rhs);
		return *this;
	}

	TextString & TextString::operator += (const TextString & rhs)
	{
		this->RakNet::RakString::operator += (rhs.Raw ());
		return *this;
	}

	bool TextString::operator == (const char * rhs) const
	{
		return ((*(RakNet::RakString *) this) == rhs);
	}
	
	bool TextString::operator == (const TextString & rhs) const
	{
		return ((*(RakNet::RakString *) this) == (const RakNet::RakString &) rhs);
	}

	bool TextString::operator != (const char * rhs) const
	{
		return ((*(RakNet::RakString *) this) != rhs);
	}

	bool TextString::operator != (const TextString & rhs) const
	{
		return ((*(RakNet::RakString *) this) != (const RakNet::RakString &) rhs);
	}

	bool TextString::ValidateChar (VALIDATETYPES type, char c)
	{
		bool isAllowed = false;
		switch (type)
		{
		case VT_NUMBER:
			if (c >= SDLK_0 && c <= SDLK_9)
			{
				isAllowed = true;
			}
			break;
		
		case VT_IP:
			if ((c >= SDLK_0 && c <= SDLK_9) || c == SDLK_PERIOD)
			{
				isAllowed = true;
			}
			break;

		case VT_ALPHANUMERIC:
			if ((c >= SDLK_0 && c <= SDLK_9) || 
				(c >= SDLK_a && c <= SDLK_z) ||
				(c >= 'A' && c <= 'Z'))
			{
				isAllowed = true;
			}
			break;

		case VT_FILENAME:
			if ((c >= SDLK_0 && c <= SDLK_9) || 
				(c >= SDLK_a && c <= SDLK_z) ||
				(c >= 'A' && c <= 'Z') ||
				c == SDLK_UNDERSCORE ||
				c == SDLK_SPACE)
			{
				isAllowed = true;
			}
			break;

		case VT_FILEPATH:
			if ((c >= SDLK_0 && c <= SDLK_9) || 
				(c >= SDLK_a && c <= SDLK_z) ||
				(c >= 'A' && c <= 'Z') ||
				c == SDLK_PERIOD ||
				c == SDLK_SLASH ||
				c == SDLK_UNDERSCORE ||
				c == SDLK_SPACE)
			{
				isAllowed = true;
			}
			break;

		case VT_TEXT:
			if ((c >= SDLK_0 && c <= SDLK_9) || 
				(c >= SDLK_a && c <= SDLK_z) ||
				(c >= 'A' && c <= 'Z') ||
				c == SDLK_UNDERSCORE)
			{
				isAllowed = true;
			}
			break;

		default:
			assert (false);
		}
		return isAllowed;
	}

	TextString TextString::Validate (VALIDATETYPES type, const char * in, char replaceWith)
	{
		assert (in != NULL);

		static char buffer [256];
		int idx = 0;
		for (; in [idx] != '\0'; ++idx)
		{
			buffer [idx] = (ValidateChar (type, in [idx])) ? in [idx] : replaceWith;
		}
		buffer [idx] ='\0';
		return TextString (buffer);
	}

#endif

}
